/*
 * Touch Pot Interface/Test program
 *
 * This hack of a sketch provides a simple interface to the Touch Pot supporting
 * both the serial and I2C interfaces.  It is essentially a repeater/translater
 * with some testing functionality and basic touchpad calibration thrown in.
 *
 * Commands using capital letters are sent through he serial interface.
 * Commands using lower-case letters are sent through the I2C interface.
 *
 * Designed to be run on a Arduino Leonardo because it has two serial
 * interfaces.
 *
 * Wiring Diagram
 *    Leonardo               Touch Pot
 *   ------------------------------------
 *      5V                     5V + V+ (note: Analog power needs to be connected)
 *     GND                     GND + V-
 *      D0(RX)                 TX
 *      D1(TX)                 RX
 *      D2(SDA)                SDA
 *      D3(SCL)                SCL
 *
 *
 * Note: Strings shortened to conserve SRAM
 *
 * Version 2 (new calibration algorithm)
 *
 * Copyright (c) 2014-2015 Dan Julio
 * This software comes with absolutely no warranty.
 * Released with the MIT license
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:

 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.

 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#include "Wire.h"

#define MAX_CMD_ARGS 5
#define CMD_TERM_CHAR '\r'
#define CMD_IGNORED_TERM '\n'

//
// Serial command processor state
//
#define RX_STATE_IDLE 0
#define RX_STATE_WAIT_ARG 1
#define RX_STATE_BUILD_ARG 2
#define RX_STATE_WAIT_END 3

//
// Number of EERPOM configuration values
//
#define NUM_EEP_VALS 18

//
// Emergency Factory Default values
//
#define FD_CONTROL_REG    0x40
#define FD_I2C_ADDR       0x08
#define FD_S1_MIN_PO      3550
#define FD_S2_MIN_PO      3650
#define FD_S1_VALID_DELTA 128
#define FD_S2_VALID_DELTA 120
#define FD_IDLE_DELTA     16
#define FC_EEP_ADDR_12    6
#define FD_GUARD_BAND     0
#define FC_EEP_ADDR_14    0xC4
#define FC_EEP_ADDR_15    0x86
#define FC_EEP_ADDR_16    0x88
#define FC_EEP_ADDR_17    0x16

//
// Number of samples to make when calibrating the touchpad
//
#define NUM_CAL_SMPLS  16
//
// Fraction of sensor minimum delta value to subract from idle to compute minimum power-on count
//
#define PO_DELTA_FRAC  0.33
//
// Minimum delta from idle for a valid touch
//
#define MIN_TOUCH_DELTA 256
//
// Maximum touch variation for a set of samples to be determined to be one valid touch
// Must see NUM_CAL_SMPLS values within this much of each other to decide we have a consistent touch
//
#define MAX_TOUCH_VARIATION 64
 

char c;
int rxState = RX_STATE_IDLE;
int argIndex = 0;
int tempArg;

unsigned int argBuf[MAX_CMD_ARGS];
int numArgs = 0;

unsigned char i2cAddr = 0x08;

byte i2cStatus;

// Globals for calibration
uint16_t curSns1Val;
uint16_t curSns2Val;
uint16_t sns1Samples[NUM_CAL_SMPLS];
uint16_t sns2Samples[NUM_CAL_SMPLS];
uint16_t sns1Avg;
uint16_t sns2Avg;

// Factory default value array - matches Flash Memory storage address locations
uint8_t fdVals[NUM_EEP_VALS] = {
  FD_CONTROL_REG,
  FD_I2C_ADDR,
  FD_S1_MIN_PO >> 8,
  FD_S1_MIN_PO & 0xFF,
  FD_S2_MIN_PO >> 8,
  FD_S2_MIN_PO & 0xFF,
  FD_S1_VALID_DELTA >> 8,
  FD_S1_VALID_DELTA & 0xFF,
  FD_S2_VALID_DELTA >> 8,
  FD_S2_VALID_DELTA & 0xFF,
  FD_IDLE_DELTA >> 8,
  FD_IDLE_DELTA & 0xFF,
  FC_EEP_ADDR_12,
  FD_GUARD_BAND,
  FC_EEP_ADDR_14,
  FC_EEP_ADDR_15,
  FC_EEP_ADDR_16,
  FC_EEP_ADDR_17
};


void setup()
{
  Serial.begin(9600); // Host
  Serial1.begin(9600); // Touch Pot
  Wire.begin();
}

void loop()
{
  // Process host serial data and look for complete commands
  if (ProcessHostCommands()) {
    switch (argBuf[0]) {
      case 'G':
        // Get CurPotValue via SIF
        ProcessSifCommand();
        break;
        
      case 'g':
        // Get CurPotValue via I2C
        GetI2cData(i2cAddr, 1);
        break;
        
      case 'l':
        // Special command to load CurPotValue directly via I2C
        Wire.beginTransmission(i2cAddr);
        Wire.write(argBuf[1]);
        i2cStatus = Wire.endTransmission();
        (void) AnalyzeI2cStatus("Direct write");
        break;
        
      case 'R':
        // Read Register via SIF
        ProcessSifCommand();
        break;
        
      case 'r':
        // Read Register via I2C
        ProcessI2cCommand();
        GetI2cData(i2cAddr+1, 1);
        break;
        
      case 'W':
        // Write Register via SIF
        ProcessSifCommand();
        break;
        
      case 'w':
        // Read Register via SIF
        ProcessI2cCommand();
        break;
        
      case 'Q':
        // Query EEPROM via SIF
        ProcessSifCommand();
        break;
        
      case 'q':
        // Query EEPROM via I2C
        ProcessI2cCommand();
        GetI2cData(i2cAddr+1, 1);
        break;
        
      case 'S':
        // Set EEPROM via SIF
        ProcessSifCommand();
        break;
        
      case 's':
        // Set EEPROM via I2C
        ProcessI2cCommand();
        break;
        
      case 'H':
        // Hammer the device via serial
        HammerSerial();
        break;
        
      case 'h':
        // Hammer the device via I2C port
        HammerI2c();
        break;
        
      case 'a':
        // Animate the LED display via I2C
        AnimateDisplay();
        break;
        
      case 'B':
        // Bad command via serial
        ProcessSifCommand();
        break;
        
      case 'b':
        // Bad command via I2C
        ProcessI2cCommand();
        break;
        
      case 'c':
        // Run the calibration routine via I2C
        CalTouchpad();
        break;
        
      case 'd':
        // Display touchpad registers
        DisplayRegisters();
        break;
        
      case 'F':
        // Restore factory defaults via serial
        RestoreDefaults();
        break;
        
      case 'I':
        // Set I2C address
        i2cAddr = argBuf[1];
        break;
        
      case 'i':
        // Print I2C address
        Serial.println(i2cAddr, DEC);
        break;
        
      case '?':
        PrintHelp();
        break;
        
      default:
        Serial.println("Unknown cmd");
    }
  }
}


// Echo any serial data from the device back to the host
void serialEvent1() {
  while (Serial1.available()) {
    c = Serial1.read();
    Serial.write(c);
  }    
}


void PrintHelp() {
  Serial.println("G: Get CPV - Ser");
  Serial.println("g: Get CPV - I2C");
  Serial.println("l <N>: CPV = <N> - I2C");
  Serial.println("R <N>: Read reg <N> - Ser");
  Serial.println("r <N>: Read reg <N> - I2C");
  Serial.println("W <N> <M>: Write reg <N> = <M> - Ser");
  Serial.println("w <N> <M>: Write reg <N> = <M> - I2C");
  Serial.println("Q <N>: Get EEP <N> - Ser");
  Serial.println("q <N>: Get EEP <N> - I2C");
  Serial.println("S <N> <M>: Set EEP <N> = <M> - Ser");
  Serial.println("s <N> <M>: Set EEP <N> = <M> - I2C");
  Serial.println("H: Hammer - Ser");
  Serial.println("h: Hammer - I2C");
  Serial.println("a: Animate LEDs");
  Serial.println("B [args]: Bad cmd - Ser");
  Serial.println("b [args]: Bad cmd - I2C");
  Serial.println("c: Cal tp");
  Serial.println("d: Disp tp info");
  Serial.println("F: Reload Fac Dflts");
  Serial.println("I <N>: Set I2C addr");
  Serial.println("i: Disp I2C addr");
  Serial.println("?: Help");
}


void PrintDone() {
  // Gotta save every byte of SRAM... (yes, I learned this the hard way...)
  Serial.println("done");
}


boolean ProcessHostCommands() {
  boolean valid = false;
  
  while (Serial.available()) {
    c = Serial.read();
    switch (rxState) {
      case RX_STATE_IDLE:
        // Look for a letter to validate the start of a command
        if (((c >= 'a') && (c <= 'z')) || ((c >= 'A') && (c <= 'Z')) || (c == '?')) {
          argBuf[0] = c;
          numArgs = 1;
          argIndex = 1;
          rxState = RX_STATE_WAIT_ARG;
        }
        break;
        
      case RX_STATE_WAIT_ARG:
        // skip white space, look for a number to start an argument
        if (((c >= 'a') && (c <= 'z')) || ((c >= 'A') && (c <= 'Z')) || (c == '?')) {
          // Let them restart a command right in the middle...
          argBuf[0] = c;
          numArgs = 1;
          argIndex = 1;
          rxState = RX_STATE_WAIT_ARG;
        } else if (c == CMD_TERM_CHAR) {
          valid = true;
          rxState = RX_STATE_IDLE;
        } else if ((c >= '0') && (c <= '9')) {
          // Start building an argument
          tempArg = c - '0';
          rxState = RX_STATE_BUILD_ARG;
        } else if ((c == ' ') || (c == CMD_IGNORED_TERM)) {
          rxState = RX_STATE_WAIT_ARG;
        } else {
          rxState = RX_STATE_WAIT_END;
        }
        break;
        
      case RX_STATE_BUILD_ARG:
        if (((c >= 'a') && (c <= 'z')) || ((c >= 'A') && (c <= 'Z')) || (c == '?')) {
          // Let them restart a command right in the middle...
          argBuf[0] = c;
          numArgs = 1;
          argIndex = 1;
          rxState = RX_STATE_WAIT_ARG;
        } else if (c == CMD_TERM_CHAR) {
          // Push the final argument
          if (argIndex < MAX_CMD_ARGS) {
            argBuf[argIndex] = tempArg;
            numArgs++;
            valid = true;
            rxState = RX_STATE_IDLE;
          } else {
            rxState = RX_STATE_WAIT_END;
          }
        } else if ((c >= '0') && (c <= '9')) {
          // Continue building an argument
          tempArg = (tempArg * 10) + (c - '0');
          rxState = RX_STATE_BUILD_ARG;
        } else if ((c == ' ') || (c == CMD_IGNORED_TERM)) {
          // Push the argument
          if (argIndex < MAX_CMD_ARGS) {
            argBuf[argIndex++] = tempArg;
            numArgs++;
            rxState = RX_STATE_WAIT_ARG;
          } else {
            rxState = RX_STATE_WAIT_END;
          }
        } else {
          rxState = RX_STATE_WAIT_END;
        }
        break;
        
      case RX_STATE_WAIT_END:
        // Wait for the terminator
        if (c == CMD_TERM_CHAR) {
          rxState = RX_STATE_IDLE;
        }
        break;
        
      default:
        rxState = RX_STATE_IDLE;
    }
  }
    
  return valid;
}

// Send a command to the device via its serial interface
void ProcessSifCommand() {
  int i;
  
  Serial1.write(argBuf[0]);
  Serial1.print(' ');
  for (i=1; i<numArgs; i++) {
    Serial1.print(argBuf[i]);
    Serial1.print(' ');
  }
  Serial1.println();
}


// Send a command to the device via its I2C interface
void ProcessI2cCommand() {
  int i;
  
  // Hack to convert the lower-case version of the ASCII command type to the upper case version the device is expecting
  argBuf[0] -= 0x20;
  Wire.beginTransmission(i2cAddr+1);
  for (i=0; i<numArgs; i++) {
    Wire.write(argBuf[i]);
  }
  i2cStatus = Wire.endTransmission();
  (void) AnalyzeI2cStatus("Ind I2C write");
}


// Read a specified number of bytes from the device via its I2C interface
void GetI2cData(uint8_t addr, uint8_t numBytes) {
  int i;
  
  Wire.requestFrom(addr, numBytes);
  if (Wire.available()) {
    for (i=0; i<numBytes; i++) {
      Serial.print(Wire.read());
    }
    Serial.println();
  } else {
    Serial.println("I2C read NACK");
  }
}


// Returns 0 for successful status, otherwise returns 1 and logs error message
byte AnalyzeI2cStatus(char *s) {
  if (i2cStatus) {
    Serial.print(s);
    Serial.print(" error: ");
    switch (i2cStatus) {
      case 1: Serial.println("I2C xmit overflow");
        break;
        
      case 2: Serial.println("Address NACK");
        break;
        
      case 3: Serial.println("Data NACK");
        break;
        
      case 4: Serial.println("Other I2C error");
        break;
        
      default: Serial.println("Unknown I2C error");
    }
    return 1;
  } else {
    return 0;
  }
}


void HammerSerial() {
  int i;
  uint8_t setVal;

  // Write increasing values, wrapping and ending at 0x55 (
  setVal = 0;
  for (i=0; i<(256 + 0x55 + 1); i++) {
    Serial1.print("W1 "); // Write Register 1 (CurPotValue)
    Serial1.println(setVal++);
  }
  
  // Perform 20 reads (whose result should fit in the arduino rx buffer, assuming it is 64 bytes)
  Serial.println("Verify 20 lines of \"85\" below");
  for (i=0; i<20; i++) {
    Serial1.println("G");
  }
  
  PrintDone();
}


void HammerI2c() {
  int i;
  uint8_t setVal;
  uint8_t getVal;
  uint8_t i2cAddr2 = i2cAddr + 1;
  
  // Hammer direct access, occasionally pausing to verify the write data
  // End at 0xA5
  setVal = 0;
  for (i=0; i<(2*256 + 0xA5 + 1); i++) {
    Wire.beginTransmission(i2cAddr);
    Wire.write(setVal);
    i2cStatus = Wire.endTransmission();
    if (AnalyzeI2cStatus("I2C direct write")) {
      Serial.print("at i = ");
      Serial.println(i, DEC);
      break;
    }
    
    if ((setVal % 64) == 0) {
      delay(1); // Delay 1 mSec to allow the command fifo to empty (since the fast I2C read bypasses the command fifo)
      Wire.requestFrom((uint8_t) i2cAddr, (uint8_t) 1);
      if (Wire.available()) {
        getVal = Wire.read();
        if (getVal != setVal) {
          Serial.print("I2C direct write series compare failed.  Exp: 0x");
          Serial.print(setVal, HEX);
          Serial.print(" got: 0x");
          Serial.println(getVal, HEX);
          break;
        }
      } else {
        Serial.print("I2C direct write series read failed @ 0x");
        Serial.println(setVal, HEX);
        break;
      }
    }
    
    setVal++;
  }
  
  // Hammer some direct reads
  setVal--;  // Go back to last value written
  delay(1); // Allow command fifo to empty before we do the first read
  for (i=0; i<256; i++) {
    Wire.requestFrom((uint8_t) i2cAddr, (uint8_t) 1);
    if (Wire.available()) {
      getVal = Wire.read();
      if (getVal != setVal) {
        Serial.print("I2C dir read series compare failed @ i=");
        Serial.print(i);
        Serial.print(".  Exp: 0x");
        Serial.print(setVal, HEX);
        Serial.print(" got: 0x");
        Serial.println(getVal, HEX);
        break;
      }
    } else {
      Serial.print("I2C dir read series read failed @ i=");
      Serial.println(i);
      break;
    }
  }

  // Hammer indirect access, occasionally pausing to verify the write data
  // End at 0x5A
  setVal = 0;
  for (i=0; i<(2*256 + 0x5A + 1); i++) {
    Wire.beginTransmission(i2cAddr2);
    Wire.write("W");
    Wire.write(1);
    Wire.write(setVal);
    i2cStatus = Wire.endTransmission();
    if (AnalyzeI2cStatus("I2C ind write")) {
      Serial.print("at i = ");
      Serial.println(i, DEC);
      break;
    }
    
    if ((setVal % 64) == 0) {
      delay(1); // Delay 1 mSec to allow the command fifo to empty (since the fast I2C read bypasses the command fifo)
      Wire.beginTransmission(i2cAddr2);
      Wire.write('R');
      Wire.write(1);
      i2cStatus = Wire.endTransmission();
      if (AnalyzeI2cStatus("I2C ind write read cmd")) {
        Serial.print("@ i = ");
        Serial.println(i, DEC);
        break;
      }
      
      Wire.requestFrom(i2cAddr2, (uint8_t) 1);
      if (Wire.available()) {
        getVal = Wire.read();
        if (getVal != setVal) {
          Serial.print("I2C ind write series compare failed.  Exp: 0x");
          Serial.print(setVal, HEX);
          Serial.print(" got: 0x");
          Serial.println(getVal, HEX);
          break;
        }
      } else {
        Serial.print("I2C ind write series read failed @ 0x");
        Serial.println(setVal, HEX);
        break;
      }
    }
    
    setVal++;
  }
  
  // Hammer some indirect reads
  setVal--;  // Go back to last value written
  
  for (i=0; i<256; i++) {
    Wire.beginTransmission(i2cAddr2);
    Wire.write('R');
    Wire.write(1);
    i2cStatus = Wire.endTransmission();
    if (AnalyzeI2cStatus("I2C ind read")) break;
    
    Wire.requestFrom(i2cAddr2, (uint8_t) 1);
    if (Wire.available()) {
      getVal = Wire.read();
      if (getVal != setVal) {
        Serial.print("I2C ind read series compare failed @ i=");
        Serial.print(i);
        Serial.print(".  Exp: 0x");
        Serial.print(setVal, HEX);
        Serial.print(" got: 0x");
        Serial.println(getVal, HEX);
        break;
      }
    } else {
      Serial.print("I2C ind read series read failed @ i=");
      Serial.println(i);
      break;
    }
  }
  
  PrintDone();
}


void WriteI2cReg(uint8_t addr, uint8_t data) {
  Wire.beginTransmission(i2cAddr+1);
  Wire.write('W');
  Wire.write(addr);
  Wire.write(data);
  i2cStatus = Wire.endTransmission();
  (void) AnalyzeI2cStatus("WriteI2cReg");
}


uint8_t ReadI2cReg(uint8_t addr) {
  Wire.beginTransmission(i2cAddr+1);
  Wire.write('R');
  Wire.write(addr);
  i2cStatus = Wire.endTransmission();
  (void) AnalyzeI2cStatus("ReadI2cReg Reg Write");
  
  Wire.requestFrom((uint8_t) (i2cAddr + 1), (uint8_t) 1);
    if (Wire.available()) {
      return(Wire.read());
    } else {
      Serial.print("ReadI2cReg ");
      Serial.print(addr);
      Serial.println(" not avail");
      return(0);
    }
}


uint16_t GetTouchpadSample(uint8_t snsNum) {
  if (snsNum == 1) {
    return((ReadI2cReg(5) << 8) | ReadI2cReg(6));
  } else {
    return((ReadI2cReg(7) << 8) | ReadI2cReg(8));
  }
}


void AnimateDisplay() {
  uint8_t origControlReg;
  int i;
  
  // Set the display to user mode - brightness '01'
  origControlReg = ReadI2cReg(3);
  WriteI2cReg(3, 0x05);
  
  // Blank the display for a brief time
  WriteI2cReg(4, 0);
  delay(500);
  
  // Display a binary count
  for (i=0; i<256; i++) {
    WriteI2cReg(4, i & 0xFF);
    delay(10);
  }
  
  // Restore the original control register value
  delay(500);
  WriteI2cReg(3, origControlReg);
  PrintDone();
}


void DisplayRegisters() {
  uint8_t statusReg;
  uint16_t tpReg;
  
  // Status register
  statusReg = ReadI2cReg(2);
  Serial.print("Stat = 0x");
  Serial.println(statusReg,HEX);
  if (statusReg & 0x01) Serial.println(" Sns1 InitCal");
  if (statusReg & 0x02) Serial.println(" Sns2 InitCal");
  if (statusReg & 0x04) Serial.println(" Sns1 ValTouch");
  if (statusReg & 0x08) Serial.println(" Sns2 ValTouch");
  
  // Sensor 1 current value and cal values
  tpReg = GetTouchpadSample(1);
  Serial.print("Sns1 = ");
  Serial.print(tpReg);
  tpReg = (ReadEepLoc(2) << 8) | ReadEepLoc(3);
  Serial.print(" - POmin = ");
  Serial.print(tpReg);
  tpReg = (ReadEepLoc(6) << 8) | ReadEepLoc(7);
  Serial.print(", dTouch = ");
  Serial.println(tpReg);
  
  // Sensor 2 current value and cal values
  tpReg = GetTouchpadSample(2);
  Serial.print("Sns2 = ");
  Serial.print(tpReg);
  tpReg = (ReadEepLoc(4) << 8) | ReadEepLoc(5);
  Serial.print(" - POmin = ");
  Serial.print(tpReg);
  tpReg = (ReadEepLoc(8) << 8) | ReadEepLoc(9);
  Serial.print(", dTouch = ");
  Serial.println(tpReg);
}


uint8_t ReadEepLoc(uint8_t addr) {
  Wire.beginTransmission(i2cAddr+1);
  Wire.write('Q');
  Wire.write(addr);
  i2cStatus = Wire.endTransmission();
  (void) AnalyzeI2cStatus("ReadEepLoc Reg Write");
  
  Wire.requestFrom((uint8_t) (i2cAddr + 1), (uint8_t) 1);
    if (Wire.available()) {
      return(Wire.read());
    } else {
      Serial.print("ReadEepLoc ");
      Serial.print(addr);
      Serial.println(" not avail");
      return(0);
    }
}


void WriteEepLoc(uint8_t addr, uint8_t data) {
  Serial1.print("S");
  Serial1.print(addr);
  Serial1.print(" ");
  Serial1.println(data);
}


void RestoreDefaults() {
  uint8_t i;
  uint8_t backupEepVals[NUM_EEP_VALS];
  uint8_t expCheckSum;
  uint8_t curCheckSum;
  boolean badData;
  
  // Read the backup copies
  curCheckSum = 0;
  badData = true;  // Set false if we get at least some non-zero data indiating I2C is working
  for (i=0; i<NUM_EEP_VALS; i++) {
    backupEepVals[i] = ReadEepLoc(255 - i);
    curCheckSum += backupEepVals[i];
    if (backupEepVals[i] != 0) {
      badData = false;
    }
  }
  // Get the checksum
  expCheckSum = ReadEepLoc(255 - NUM_EEP_VALS);
  
  // Restore either the set from the device itself or our "emergency" set
  if ((curCheckSum != expCheckSum) || badData) {
    Serial.print("Emergency set:");
    for (i=0; i<NUM_EEP_VALS; i++) {
      WriteEepLoc(i, fdVals[i]);
      Serial.print(".");
    }
  } else {
    Serial.print("Device set:");
    for (i=0; i<NUM_EEP_VALS; i++) {
      WriteEepLoc(i, backupEepVals[i]);
      Serial.print(".");
    }
  }
  PrintDone();
  Serial.println("Reset hw");
}


// CalTouchpad requires the user to touch at each end and computes the Mininum power-on and delta
// touch values.  It assumes default values for the other Touch Sensor configuration items.
// The user should really be touching the very end of the sensor at both touches or the calibration
// will be off.
void CalTouchpad() {
  uint8_t origControlReg;
  uint8_t ledPattern;

  uint16_t sns1IdleAvg;
  uint16_t sns2IdleAvg;
  
  uint16_t sns1MinPO;
  uint16_t sns2MinPO;
  uint16_t sns1MinDelta;
  uint16_t sns2MinDelta;
  
  uint16_t t;
  int numValSamples;
  int i;
  
  // Initialize the calibration.  The touchpad must start out with no capacitive load (untouched).
  // Give the user some warning and indicate it by stepping the LEDS down 5..4..3..2..1..<all off>
  origControlReg = ReadI2cReg(3); // Save the original control reg values
  WriteI2cReg(3, 0x04);  // Configure for user display, low intensity
  ledPattern = 0x10;
  WriteI2cReg(4, ledPattern);
  Serial.println("Cal in 5 secs, don't touch!");
  for (i=0; i<5; i++) {
    delay(1000);
    ledPattern >>= 1;
    WriteI2cReg(4, ledPattern);
  }
  delay(2000);
  
  // Get a small average for the untouched value
  GetIdleAverage();
  sns1IdleAvg = sns1Avg;
  sns2IdleAvg = sns2Avg;
  Serial.print(" Sns1 Idle = ");
  Serial.println(sns1Avg);
  Serial.print(" Sns2 Idle = ");
  Serial.println(sns2Avg);
  
  // Have the user touch sensor 1 (the lower sensor) and collect some readings to average.
  // When touching the sensor, make sure the finger is only over the "landing pad".  It shouldn't
  // be touching any of the teeth of the other sensor.  We want the other sensor as unaffected
  // by this touch as possible (although it will be picking up some capacitive load).
  Serial.println("Touch only lower sensor");
  BlinkLeds(0x01);
  numValSamples = 0;
  i = 0;
  while ((numValSamples < NUM_CAL_SMPLS) && (i < 128)) {
    curSns1Val = GetTouchpadSample(1);
    curSns2Val = GetTouchpadSample(2);
    if (curSns1Val <= (sns1IdleAvg - MIN_TOUCH_DELTA)) {
      if (ValidateSample(curSns1Val, sns1Samples, numValSamples)) {
        sns1Samples[numValSamples] = curSns1Val;
        sns2Samples[numValSamples] = curSns2Val;
        numValSamples++;
      } else {
        numValSamples = 0;
      }
    }
    i++;
    delay(40);  // Let the touchpad have computed a new value
  }
  if (i == 256) {
    Serial.println("Bad touch 1");
    goto RestoreDevice;
  }
  ComputeAverages(NUM_CAL_SMPLS);
  sns2MinDelta = sns2IdleAvg - sns2Avg;
  Serial.print(" Sns2 DeltaTouch = ");
  Serial.println(sns2MinDelta);
  // Rough validation that sns2MinDelta < 0.25 * (touch delta for sensor 1)
  t = sns1IdleAvg - sns1Avg;
  if (sns2MinDelta > (t >> 2)) {
    Serial.println(" Too high");
    goto RestoreDevice;
  }
  
  // Now do the same for touch sensor 2.
  Serial.println("Touch only upper sensor");
  BlinkLeds(0x80);
  numValSamples = 0;
  i = 0;
  while ((numValSamples < NUM_CAL_SMPLS) && (i < 128)) {
    curSns1Val = GetTouchpadSample(1);
    curSns2Val = GetTouchpadSample(2);
    if (curSns2Val <= (sns2IdleAvg - MIN_TOUCH_DELTA)) {
      if (ValidateSample(curSns2Val, sns2Samples, numValSamples)) {
        sns1Samples[numValSamples] = curSns1Val;
        sns2Samples[numValSamples] = curSns2Val;
        numValSamples++;
      } else {
        numValSamples = 0;
      }
    }
    i++;
    delay(40);  // Let the touchpad have computed a new value
  }
  if (i == 256) {
    Serial.println("Bad touch 2");
    goto RestoreDevice;
  }
  ComputeAverages(NUM_CAL_SMPLS);
  sns1MinDelta = sns1IdleAvg - sns1Avg;
  Serial.print(" Sns1 DeltaTouch = ");
  Serial.println(sns1MinDelta);
  // Rough validation that sns1MinDelta < 0.25 * (touch delta for sensor 2)
  t = sns2IdleAvg - sns2Avg;
  if (sns1MinDelta > (t >> 2)) {
    Serial.println(" Too high");
    goto RestoreDevice;
  }
  
  // Finally, compute the minimum power on values.  This is the idle value minus PO_DELTA_FRAC * the
  // sensor Min Delta. 
  sns1MinPO = sns1IdleAvg - round(PO_DELTA_FRAC * sns1MinDelta);
  sns2MinPO = sns2IdleAvg - round(PO_DELTA_FRAC * sns2MinDelta);
  Serial.print("Sns1 POmin = ");
  Serial.println(sns1MinPO);
  Serial.print("Sns2 POmin = ");
  Serial.println(sns2MinPO);
  
  // Load the calibration values if we've made it here
  WriteEepLoc(2, sns1MinPO >> 8);
  WriteEepLoc(3, sns1MinPO & 0xFF);
  WriteEepLoc(4, sns2MinPO >> 8);
  WriteEepLoc(5, sns2MinPO & 0xFF);
  WriteEepLoc(6, sns1MinDelta >> 8);
  WriteEepLoc(7, sns1MinDelta & 0xFF);
  WriteEepLoc(8, sns2MinDelta >> 8);
  WriteEepLoc(9, sns2MinDelta & 0xFF);
  
  PrintDone();
    
  // Jump point for code that bails on the calibration
  RestoreDevice:
  // Finally restore the original control reg value
  WriteI2cReg(3, origControlReg);
  WriteI2cReg(1, 0); // set CurPotValue = 0
}


void GetIdleAverage() {
  int i;
  
  // Get some samples
  for (i=0; i<8; i++) {
    curSns1Val = GetTouchpadSample(1);
    curSns2Val = GetTouchpadSample(2);
    sns1Samples[i] = curSns1Val;
    sns2Samples[i] = curSns2Val;
    delay(40);  // Let the touchpad have computed a new value
  }
  
  // Average them
  ComputeAverages(8);  
}


void ComputeAverages(uint8_t num) {
  uint8_t i;
  uint32_t sns1accum = 0;
  uint32_t sns2accum = 0;
  
  for (i=0; i<num; i++) {
    sns1accum += sns1Samples[i];
    sns2accum += sns2Samples[i];
  }
  
  sns1Avg = sns1accum / num;
  if ((sns1accum % num) > (num >> 1)) {
    // Round up
    sns1Avg++;
  }
  sns2Avg = sns2accum / num;
  if ((sns2accum % num) > (num >> 1)) {
    // Round up
    sns2Avg++;
  }
}

boolean ValidateSample(uint16_t curVal, uint16_t* smpArray, int numValid) {
  int i;
  uint16_t indexedVal;
  
  if (numValid == 0) {
    // First sample
    return(true);
  } else {
    for (i=0; i<numValid; i++) {
      indexedVal = *(smpArray + i);
      if (((curVal > indexedVal) && ((curVal - indexedVal) > MAX_TOUCH_VARIATION)) ||
          ((curVal <= indexedVal) && ((indexedVal - curVal) > MAX_TOUCH_VARIATION))) {
        // Sample out of range of current series
        return(false);
      }
    }
    return(true);
  }
}

void BlinkLeds(uint8_t p) {
  int i;
  
  for (i=0; i<5; i++) {
    WriteI2cReg(4, p);
    delay(500);
    WriteI2cReg(4, 0);
    delay(500);
  }
}
